import array
import struct

from cryptography.hazmat.primitives.ciphers.algorithms import AES


def key_string_to_bytes(key_string):
    num = int(key_string, 16)
    return struct.pack(">QQ", num >> 64, num & 0xFFFFFFFFFFFFFFFF)


def compute_encryption_vector(data_length, address, counter):
    vector = bytes(0)
    for i in range(data_length // AES.block_size):
        vector += struct.pack("<BLBLLBB", 0x01, 0x00, 0x00, address, counter, 0, i + 1)
    return vector


def xor_bytes(first, second):
    return [
        data_block ^ s_block
        for data_block, s_block in zip(
            array.array("B", first), array.array("B", second)
        )
    ]


def waspmote_power_to_dbm(power_index):
    """Convert Libelium Waspmote power index to dBm"""
    return 17 - (3 * power_index)



import base64
import struct
from enum import IntEnum

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import ECB

from .util import compute_encryption_vector, key_string_to_bytes, xor_bytes

JOIN_ACCEPT_DELAY1 = 5 * 1000 * 1000  # microseconds


class FrameControl:
    def __init__(
        self,
        adr: bool = False,
        adr_ack_req: bool = False,
        ack: bool = False,
        options_len: int = 0,
    ):
        self.adr = adr
        self.adr_ack_req = adr_ack_req
        self.ack = ack
        self.options_len = options_len

    @classmethod
    def loads(cls, frame_ctrl: int):
        try:
            adr = bool(frame_ctrl & 0b10000000)
            adr_ack_req = bool(frame_ctrl & 0b01000000)
            ack = bool(frame_ctrl & 0b00100000)
            options_length = frame_ctrl & 0b00000111
        except Exception:
            raise TypeError("Corrupted frame control field")

        return cls(adr, adr_ack_req, ack, options_length)


class LorawanType(IntEnum):
    JOIN_REQUEST = 0b000
    JOIN_ACCEPT = 0b001
    UNCONFIRMED_DATA_UP = 0b010
    UNCONFIRMED_DATA_DOWN = 0b011
    CONFIRMED_DATA_UP = 0b100
    CONFIRMED_DATA_DOWN = 0b101


class LorawanMessage:
    LORAWAN_VERSION_1 = 0

    def __init__(
        self,
        lorawan_type: int,
        version: int,
        mic,
        address: str,
        counter: int,
        control: FrameControl,
        options: str,
        port: int,
        payload: str,
    ):
        self.lorawan_type: int = lorawan_type
        self.version: int = version
        self.mic = mic
        self.address: str = address
        self.counter: int = counter
        self.control: FrameControl = control  # declared in frame_control field
        self.options: str = options
        self.port: int = port
        self.payload: str = payload
        self.original_len: int = len(payload)

    @classmethod
    def deserialize(cls, data, app_key=None):
        (
            lorawan_type,
            lorawan_version,
            mac_layer_payload,
            mic,
        ) = cls.__deserialize_mac_layer(data)
        (
            dev_address,
            frame_cnt,
            frame_ctrl,
            options,
            port,
            payload,
        ) = cls.__deserialize_frame_layer(mac_layer_payload)

        message = cls(
            lorawan_type,
            lorawan_version,
            mic,
            dev_address,
            frame_cnt,
            frame_ctrl,
            options,
            port,
            payload,
        )

        if app_key is not None:
            # serialize / deserialize payload
            message.payload = message.encrypt(app_key)

        return message

    @staticmethod
    def __deserialize_mac_layer(r_data):
        data = str(r_data)
        data += "=" * (4 - len(data) % 4)  # add padding to be multiple of 4
        # res = base64.urlsafe_b64decode(data)
        res = base64.b64decode(data)
        mac_header = res[0]
        lorawan_type = (mac_header & 0b11100000) >> 5
        lorawan_version = mac_header & 0b00000011
        payload = res[1 : len(res) - 4]
        mic = res[len(res) - 4 :]
        return lorawan_type, lorawan_version, payload, mic

    @staticmethod
    def __deserialize_frame_layer(data):
        dev_address, frame_ctrl, frame_cnt = struct.unpack("<LBH", data[:7])
        frame_ctrl = FrameControl.loads(frame_ctrl)
        options = data[7 : frame_ctrl.options_len] if frame_ctrl.options_len > 0 else ""

        if len(data[7 + frame_ctrl.options_len :]) > 0:
            port = data[7 + frame_ctrl.options_len]
            payload = data[8 + frame_ctrl.options_len :]
        else:
            port = -1
            payload = ""

        return (
            dev_address,
            frame_cnt,
            frame_ctrl,
            options,
            port,
            payload,
        )

    def encrypt(self, key_string):
        data = self.payload
        key = key_string_to_bytes(key_string)

        data += b"\0" * (AES.block_size - len(data) % AES.block_size)  # add padding
        cypher = Cipher(AES(key), ECB(), backend=default_backend()).encryptor()

        vector = compute_encryption_vector(len(data), self.address, self.counter)
        s = cypher.update(vector) + cypher.finalize()

        return xor_bytes(data, s)

    def decrypt(self, key_string):
        """Decryption is symmetrical to encryption"""
        return self.encrypt(key_string)



import struct
from enum import IntEnum

GWMP_MESSAGE_FORMAT = ">BHB"
GATEWAY_ADDRESS_FORMAT = ">Q"


class GatewayMessageType(IntEnum):
    PUSH_DATA = 0x00
    PUSH_ACK = 0x01
    PULL_DATA = 0x02
    PULL_ACK = 0x04
    PULL_RESP = 0x03
    TX_ACK = 0x05


class GatewayMessage:
    def __init__(self, version, token, type, gateway, payload):
        self.version = version
        self.token = token
        self.type = type
        self.gateway = gateway
        self.payload = payload

    @classmethod
    def from_gateway_message(cls, data):
        """Build GWMP message"""
        size = struct.calcsize(GWMP_MESSAGE_FORMAT)
        version, token, gwmp_type = struct.unpack(GWMP_MESSAGE_FORMAT, data[:size])
        if gwmp_type in (
            GatewayMessageType.TX_ACK,
            GatewayMessageType.PULL_RESP,
            GatewayMessageType.PULL_ACK,
        ):
            gateway = 0
        else:
            gateway = struct.unpack(GATEWAY_ADDRESS_FORMAT, data[size : size + 8])[0]
        payload = data[size + 8 :]
        return cls(version, token, gwmp_type, gateway, payload)



from pathlib import Path

from setuptools import setup

project_dir = Path(__file__).parent.absolute()

with open(project_dir / "requirements.txt") as f:
    install_requires = f.read().splitlines()

with open(project_dir / "README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="gwmp",
    packages=["gwmp"],
    scripts=["bin/gwmp"],
    version="0.0.1",
    description="Parser for GWMP messages used in LoRa and LoRaWAN networks",
    author="Alessio Sanfratello",
    url="https://github.com/alesanfra/lora-gwmp",
    install_requires=install_requires,
    python_requires=">=3.6",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
        "Intended Audience :: Telecommunications Industry",
        "Intended Audience :: Science/Research",
    ],
)
